import 'dart:collection';
import 'dart:convert';
import 'dart:io';

import 'package:build/build.dart';
import 'package:fi18n/src/fi18n_impl.dart';
import 'package:fi18n/src/models/locale.dart';

Builder arbToDartBuilder(BuilderOptions options) => ArbToDartBuilder(options);

Builder postBuilder(BuilderOptions options) => PostBuilder();

final _locales = <Locale,bool>{};
String _fi18nFileContent;
final _generatedLocalesDartFiles = <String>[];
String defaultLocale = 'en';

class ArbToDartBuilder implements Builder {
  ArbToDartBuilder(this.options);

  final BuilderOptions options;

  @override
  final buildExtensions = const {
    '.arb': ['.fi18n']
  };

  @override
  Future<Function> build(BuildStep buildStep) async {
    var inputId = buildStep.inputId;

    final filePath = inputId.path.substring(0, inputId.path.lastIndexOf('/'));
    final defaultArbFilePathDir = 'lib/res';

    if (filePath != defaultArbFilePathDir) {
      throw Exception('Invalid .arb file path. add your .arb files in $defaultArbFilePathDir');
    }

    String generatedClassName = 'S';
    defaultLocale = options.config['default_locale'];

    final fileMeta = generateArbFileMetaFromArb(
      filename: inputId.pathSegments.last,
      defaultLocaleName: defaultLocale
    );

    _locales[fileMeta.locale] = fileMeta.isDefault;

    if (fileMeta.isDefault) {
      final sClassContent = generateSClass(
        defaultArbContent: await buildStep.readAsString(inputId),
        generatedClassName: generatedClassName,
      );

      _fi18nFileContent = sClassContent;
      final languageDartFile = inputId.changeExtension('.fi18n');
      await buildStep.writeAsString(languageDartFile, sClassContent);
    }

    // .arb file content
    var arbContent = await buildStep.readAsString(inputId);

    final messages = json.decode(arbContent) as LinkedHashMap<String, dynamic>;

    final dartContent = generateLanguageScript(
      messages,
      locale: fileMeta.locale,
      isDefaultLocaleFile: fileMeta.isDefault,
      generatedClassName: generatedClassName,
    );

    // save to use later
    _generatedLocalesDartFiles.add(dartContent);
    return Future.value();
  }
}

class PostBuilder implements Builder {

  @override
  final buildExtensions = const {
    '.fi18n': ['.dart']
  };

  @override
  Future<Function> build(BuildStep buildStep) async {
    var exist = false;
    _locales.keys.forEach((locale) {
      if (locale.toString() == defaultLocale) {
        exist = true;
      }
    });

    if (!exist) {
      throw Exception('Default locale file not found.');
    }

    String generatedClassName = 'S';
    final content = StringBuffer();

    content
      ..writeln('')
      ..writeln('// GENERATED CODE, do not edit this file.')
      ..writeln('// This file was generated by Fi18n library. ')
      ..writeln('')
      ..writeln("import 'package:flutter/material.dart';")
      ..writeln("import 'package:flutter/foundation.dart';")
      ..writeln('')
      ..writeln(_fi18nFileContent);
    _generatedLocalesDartFiles.forEach((localeDartFile) {
      content.writeln(localeDartFile);
    });

    final delegateContent = generateDelegateScript(
      objectClassName: generatedClassName,
      supportedLocales: _locales.keys.toList()
    );

    content.writeln(delegateContent);

    final fileName = 'fi18n.dart';
    final filePath = 'lib/res';
    final file = File('$filePath/$fileName');

    if (file.existsSync()) {
      await file.delete();
    }

    await file.writeAsString(content.toString(), mode: FileMode.write);
    file.create(recursive: true);
    return Future.value();
  }
}